@@ -29,9 +29,11 @@
 #
 # See the README file for information on usage and redistribution.
 #
+from __future__ import annotations
 
 import math
 import numbers
+import struct
 
 from . import Image, ImageColor
 
@@ -44,6 +46,8 @@ directly.
 
 
 class ImageDraw:
+    font = None
+
     def __init__(self, im, mode=None):
         """
         Create a drawing instance.
@@ -65,11 +69,13 @@ class ImageDraw:
             if mode == "RGBA" and im.mode == "RGB":
                 blend = 1
             else:
-                raise ValueError("mode mismatch")
+                msg = "mode mismatch"
+                raise ValueError(msg)
         if mode == "P":
             self.palette = im.palette
         else:
             self.palette = None
+        self._image = im
         self.im = im.im
         self.draw = Image.core.draw(self.im, blend)
         self.mode = mode
@@ -82,13 +88,25 @@ class ImageDraw:
             self.fontmode = "1"
         else:
             self.fontmode = "L"  # aliasing is okay for other modes
-        self.fill = 0
-        self.font = None
+        self.fill = False
 
     def getfont(self):
         """
         Get the current default font.
 
+        To set the default font for this ImageDraw instance::
+
+            from PIL import ImageDraw, ImageFont
+            draw.font = ImageFont.truetype("Tests/fonts/FreeMono.ttf")
+
+        To set the default font for all future ImageDraw instances::
+
+            from PIL import ImageDraw, ImageFont
+            ImageDraw.ImageDraw.font = ImageFont.truetype("Tests/fonts/FreeMono.ttf")
+
+        If the current default font is ``None``,
+        it is initialized with ``ImageFont.load_default()``.
+
         :returns: An image font."""
         if not self.font:
             # FIXME: should add a font repository
@@ -97,6 +115,15 @@ class ImageDraw:
             self.font = ImageFont.load_default()
         return self.font
 
+    def _getfont(self, font_size):
+        if font_size is not None:
+            from . import ImageFont
+
+            font = ImageFont.load_default(font_size)
+        else:
+            font = self.getfont()
+        return font
+
     def _getink(self, ink, fill=None):
         if ink is None and fill is None:
             if self.fill:
@@ -108,13 +135,13 @@ class ImageDraw:
                 if isinstance(ink, str):
                     ink = ImageColor.getcolor(ink, self.mode)
                 if self.palette and not isinstance(ink, numbers.Number):
-                    ink = self.palette.getcolor(ink)
+                    ink = self.palette.getcolor(ink, self._image)
                 ink = self.draw.draw_ink(ink)
             if fill is not None:
                 if isinstance(fill, str):
                     fill = ImageColor.getcolor(fill, self.mode)
                 if self.palette and not isinstance(fill, numbers.Number):
-                    fill = self.palette.getcolor(fill)
+                    fill = self.palette.getcolor(fill, self._image)
                 fill = self.draw.draw_ink(fill)
         return ink, fill
 
@@ -173,13 +200,11 @@ class ImageDraw:
                         angle -= 90
                         distance = width / 2 - 1
                         return tuple(
-                            [
-                                p + (math.floor(p_d) if p_d > 0 else math.ceil(p_d))
-                                for p, p_d in (
-                                    (x, distance * math.cos(math.radians(angle))),
-                                    (y, distance * math.sin(math.radians(angle))),
-                                )
-                            ]
+                            p + (math.floor(p_d) if p_d > 0 else math.ceil(p_d))
+                            for p, p_d in (
+                                (x, distance * math.cos(math.radians(angle))),
+                                (y, distance * math.sin(math.radians(angle))),
+                            )
                         )
 
                     flipped = (
@@ -198,18 +223,18 @@ class ImageDraw:
                     if width > 8:
                         # Cover potential gaps between the line and the joint
                         if flipped:
-                            gapCoords = [
+                            gap_coords = [
                                 coord_at_angle(point, angles[0] + 90),
                                 point,
                                 coord_at_angle(point, angles[1] + 90),
                             ]
                         else:
-                            gapCoords = [
+                            gap_coords = [
                                 coord_at_angle(point, angles[0] - 90),
                                 point,
                                 coord_at_angle(point, angles[1] - 90),
                             ]
-                        self.line(gapCoords, fill, width=3)
+                        self.line(gap_coords, fill, width=3)
 
     def shape(self, shape, fill=None, outline=None):
         """(Experimental) Draw a shape."""
@@ -234,20 +259,42 @@ class ImageDraw:
         if ink is not None:
             self.draw.draw_points(xy, ink)
 
-    def polygon(self, xy, fill=None, outline=None):
+    def polygon(self, xy, fill=None, outline=None, width=1):
         """Draw a polygon."""
         ink, fill = self._getink(outline, fill)
         if fill is not None:
             self.draw.draw_polygon(xy, fill, 1)
-        if ink is not None and ink != fill:
-            self.draw.draw_polygon(xy, ink, 0)
+        if ink is not None and ink != fill and width != 0:
+            if width == 1:
+                self.draw.draw_polygon(xy, ink, 0, width)
+            else:
+                # To avoid expanding the polygon outwards,
+                # use the fill as a mask
+                mask = Image.new("1", self.im.size)
+                mask_ink = self._getink(1)[0]
+
+                fill_im = mask.copy()
+                draw = Draw(fill_im)
+                draw.draw.draw_polygon(xy, mask_ink, 1)
+
+                ink_im = mask.copy()
+                draw = Draw(ink_im)
+                width = width * 2 - 1
+                draw.draw.draw_polygon(xy, mask_ink, 0, width)
+
+                mask.paste(ink_im, mask=fill_im)
+
+                im = Image.new(self.mode, self.im.size)
+                draw = Draw(im)
+                draw.draw.draw_polygon(xy, ink, 0, width)
+                self.im.paste(im.im, (0, 0) + im.size, mask.im)
 
     def regular_polygon(
-        self, bounding_circle, n_sides, rotation=0, fill=None, outline=None
+        self, bounding_circle, n_sides, rotation=0, fill=None, outline=None, width=1
     ):
         """Draw a regular polygon."""
         xy = _compute_regular_polygon_vertices(bounding_circle, n_sides, rotation)
-        self.polygon(xy, fill, outline)
+        self.polygon(xy, fill, outline, width)
 
     def rectangle(self, xy, fill=None, outline=None, width=1):
         """Draw a rectangle."""
@@ -257,8 +304,134 @@ class ImageDraw:
         if ink is not None and ink != fill and width != 0:
             self.draw.draw_rectangle(xy, ink, 0, width)
 
+    def rounded_rectangle(
+        self, xy, radius=0, fill=None, outline=None, width=1, *, corners=None
+    ):
+        """Draw a rounded rectangle."""
+        if isinstance(xy[0], (list, tuple)):
+            (x0, y0), (x1, y1) = xy
+        else:
+            x0, y0, x1, y1 = xy
+        if x1 < x0:
+            msg = "x1 must be greater than or equal to x0"
+            raise ValueError(msg)
+        if y1 < y0:
+            msg = "y1 must be greater than or equal to y0"
+            raise ValueError(msg)
+        if corners is None:
+            corners = (True, True, True, True)
+
+        d = radius * 2
+
+        full_x, full_y = False, False
+        if all(corners):
+            full_x = d >= x1 - x0 - 1
+            if full_x:
+                # The two left and two right corners are joined
+                d = x1 - x0
+            full_y = d >= y1 - y0 - 1
+            if full_y:
+                # The two top and two bottom corners are joined
+                d = y1 - y0
+            if full_x and full_y:
+                # If all corners are joined, that is a circle
+                return self.ellipse(xy, fill, outline, width)
+
+        if d == 0 or not any(corners):
+            # If the corners have no curve,
+            # or there are no corners,
+            # that is a rectangle
+            return self.rectangle(xy, fill, outline, width)
+
+        r = d // 2
+        ink, fill = self._getink(outline, fill)
+
+        def draw_corners(pieslice):
+            if full_x:
+                # Draw top and bottom halves
+                parts = (
+                    ((x0, y0, x0 + d, y0 + d), 180, 360),
+                    ((x0, y1 - d, x0 + d, y1), 0, 180),
+                )
+            elif full_y:
+                # Draw left and right halves
+                parts = (
+                    ((x0, y0, x0 + d, y0 + d), 90, 270),
+                    ((x1 - d, y0, x1, y0 + d), 270, 90),
+                )
+            else:
+                # Draw four separate corners
+                parts = []
+                for i, part in enumerate(
+                    (
+                        ((x0, y0, x0 + d, y0 + d), 180, 270),
+                        ((x1 - d, y0, x1, y0 + d), 270, 360),
+                        ((x1 - d, y1 - d, x1, y1), 0, 90),
+                        ((x0, y1 - d, x0 + d, y1), 90, 180),
+                    )
+                ):
+                    if corners[i]:
+                        parts.append(part)
+            for part in parts:
+                if pieslice:
+                    self.draw.draw_pieslice(*(part + (fill, 1)))
+                else:
+                    self.draw.draw_arc(*(part + (ink, width)))
+
+        if fill is not None:
+            draw_corners(True)
+
+            if full_x:
+                self.draw.draw_rectangle((x0, y0 + r + 1, x1, y1 - r - 1), fill, 1)
+            else:
+                self.draw.draw_rectangle((x0 + r + 1, y0, x1 - r - 1, y1), fill, 1)
+            if not full_x and not full_y:
+                left = [x0, y0, x0 + r, y1]
+                if corners[0]:
+                    left[1] += r + 1
+                if corners[3]:
+                    left[3] -= r + 1
+                self.draw.draw_rectangle(left, fill, 1)
+
+                right = [x1 - r, y0, x1, y1]
+                if corners[1]:
+                    right[1] += r + 1
+                if corners[2]:
+                    right[3] -= r + 1
+                self.draw.draw_rectangle(right, fill, 1)
+        if ink is not None and ink != fill and width != 0:
+            draw_corners(False)
+
+            if not full_x:
+                top = [x0, y0, x1, y0 + width - 1]
+                if corners[0]:
+                    top[0] += r + 1
+                if corners[1]:
+                    top[2] -= r + 1
+                self.draw.draw_rectangle(top, ink, 1)
+
+                bottom = [x0, y1 - width + 1, x1, y1]
+                if corners[3]:
+                    bottom[0] += r + 1
+                if corners[2]:
+                    bottom[2] -= r + 1
+                self.draw.draw_rectangle(bottom, ink, 1)
+            if not full_y:
+                left = [x0, y0, x0 + width - 1, y1]
+                if corners[0]:
+                    left[1] += r + 1
+                if corners[3]:
+                    left[3] -= r + 1
+                self.draw.draw_rectangle(left, ink, 1)
+
+                right = [x1 - width + 1, y0, x1, y1]
+                if corners[1]:
+                    right[1] += r + 1
+                if corners[2]:
+                    right[3] -= r + 1
+                self.draw.draw_rectangle(right, ink, 1)
+
     def _multiline_check(self, text):
-        """Draw text."""
         split_character = "\n" if isinstance(text, str) else b"\n"
 
         return split_character in text
@@ -268,6 +441,13 @@ class ImageDraw:
 
         return text.split(split_character)
 
+    def _multiline_spacing(self, font, spacing, stroke_width):
+        return (
+            self.textbbox((0, 0), "A", font, stroke_width=stroke_width)[3]
+            + stroke_width
+            + spacing
+        )
+
     def text(
         self,
         xy,
@@ -286,6 +466,14 @@ class ImageDraw:
         *args,
         **kwargs,
     ):
+        """Draw text."""
+        if embedded_color and self.mode not in ("RGB", "RGBA"):
+            msg = "Embedded color supported only in RGB and RGBA modes"
+            raise ValueError(msg)
+
+        if font is None:
+            font = self._getfont(kwargs.get("font_size"))
+
         if self._multiline_check(text):
             return self.multiline_text(
                 xy,
@@ -303,12 +491,6 @@ class ImageDraw:
                 embedded_color,
             )
 
-        if embedded_color and self.mode not in ("RGB", "RGBA"):
-            raise ValueError("Embedded color supported only in RGB and RGBA modes")
-
-        if font is None:
-            font = self.getfont()
-
         def getink(fill):
             ink, fill = self._getink(fill)
             if ink is None:
@@ -319,7 +501,11 @@ class ImageDraw:
             mode = self.fontmode
             if stroke_width == 0 and embedded_color:
                 mode = "RGBA"
-            coord = xy
+            coord = []
+            start = []
+            for i in range(2):
+                coord.append(int(xy[i]))
+                start.append(math.modf(xy[i])[0])
             try:
                 mask, offset = font.getmask2(
                     text,
@@ -330,6 +516,7 @@ class ImageDraw:
                     stroke_width=stroke_width,
                     anchor=anchor,
                     ink=ink,
+                    start=start,
                     *args,
                     **kwargs,
                 )
@@ -345,6 +532,7 @@ class ImageDraw:
                         stroke_width,
                         anchor,
                         ink,
+                        start=start,
                         *args,
                         **kwargs,
                     )
@@ -356,9 +544,10 @@ class ImageDraw:
                 # font.getmask2(mode="RGBA") returns color in RGB bands and mask in A
                 # extract mask and set text alpha
                 color, mask = mask, mask.getband(3)
-                color.fillband(3, (ink >> 24) & 0xFF)
-                coord2 = coord[0] + mask.size[0], coord[1] + mask.size[1]
-                self.im.paste(color, coord + coord2, mask)
+                ink_alpha = struct.pack("i", ink)[3]
+                color.fillband(3, ink_alpha)
+                x, y = coord
+                self.im.paste(color, (x, y, x + mask.size[0], y + mask.size[1]), mask)
             else:
                 self.draw.draw_bitmap(coord, mask, ink)
 
@@ -393,23 +582,29 @@ class ImageDraw:
         stroke_width=0,
         stroke_fill=None,
         embedded_color=False,
+        *,
+        font_size=None,
     ):
         if direction == "ttb":
-            raise ValueError("ttb direction is unsupported for multiline text")
+            msg = "ttb direction is unsupported for multiline text"
+            raise ValueError(msg)
 
         if anchor is None:
             anchor = "la"
         elif len(anchor) != 2:
-            raise ValueError("anchor must be a 2 character string")
+            msg = "anchor must be a 2 character string"
+            raise ValueError(msg)
         elif anchor[1] in "tb":
-            raise ValueError("anchor not supported for multiline text")
+            msg = "anchor not supported for multiline text"
+            raise ValueError(msg)
+
+        if font is None:
+            font = self._getfont(font_size)
 
         widths = []
         max_width = 0
         lines = self._multiline_split(text)
-        line_spacing = (
-            self.textsize("A", font=font, stroke_width=stroke_width)[1] + spacing
-        )
+        line_spacing = self._multiline_spacing(font, spacing, stroke_width)
         for line in lines:
             line_width = self.textlength(
                 line, font, direction=direction, features=features, language=language
@@ -441,7 +636,8 @@ class ImageDraw:
             elif align == "right":
                 left += width_difference
             else:
-                raise ValueError('align must be "left", "center" or "right"')
+                msg = 'align must be "left", "center" or "right"'
+                raise ValueError(msg)
 
             self.text(
                 (left, top),
@@ -458,48 +654,6 @@ class ImageDraw:
             )
             top += line_spacing
 
-    def textsize(
-        self,
-        text,
-        font=None,
-        spacing=4,
-        direction=None,
-        features=None,
-        language=None,
-        stroke_width=0,
-    ):
-        """Get the size of a given string, in pixels."""
-        if self._multiline_check(text):
-            return self.multiline_textsize(
-                text, font, spacing, direction, features, language, stroke_width
-            )
-
-        if font is None:
-            font = self.getfont()
-        return font.getsize(text, direction, features, language, stroke_width)
-
-    def multiline_textsize(
-        self,
-        text,
-        font=None,
-        spacing=4,
-        direction=None,
-        features=None,
-        language=None,
-        stroke_width=0,
-    ):
-        max_width = 0
-        lines = self._multiline_split(text)
-        line_spacing = (
-            self.textsize("A", font=font, stroke_width=stroke_width)[1] + spacing
-        )
-        for line in lines:
-            line_width, line_height = self.textsize(
-                line, font, spacing, direction, features, language, stroke_width
-            )
-            max_width = max(max_width, line_width)
-        return max_width, len(lines) * line_spacing - spacing
-
     def textlength(
         self,
         text,
@@ -508,25 +662,21 @@ class ImageDraw:
         features=None,
         language=None,
         embedded_color=False,
+        *,
+        font_size=None,
     ):
         """Get the length of a given string, in pixels with 1/64 precision."""
         if self._multiline_check(text):
-            raise ValueError("can't measure length of multiline text")
+            msg = "can't measure length of multiline text"
+            raise ValueError(msg)
         if embedded_color and self.mode not in ("RGB", "RGBA"):
-            raise ValueError("Embedded color supported only in RGB and RGBA modes")
+            msg = "Embedded color supported only in RGB and RGBA modes"
+            raise ValueError(msg)
 
         if font is None:
-            font = self.getfont()
+            font = self._getfont(font_size)
         mode = "RGBA" if embedded_color else self.fontmode
-        try:
-            return font.getlength(text, mode, direction, features, language)
-        except AttributeError:
-            size = self.textsize(
-                text, font, direction=direction, features=features, language=language
-            )
-            if direction == "ttb":
-                return size[1]
-            return size[0]
+        return font.getlength(text, mode, direction, features, language)
 
     def textbbox(
         self,
@@ -541,10 +691,16 @@ class ImageDraw:
         language=None,
         stroke_width=0,
         embedded_color=False,
+        *,
+        font_size=None,
     ):
         """Get the bounding box of a given string, in pixels."""
         if embedded_color and self.mode not in ("RGB", "RGBA"):
-            raise ValueError("Embedded color supported only in RGB and RGBA modes")
+            msg = "Embedded color supported only in RGB and RGBA modes"
+            raise ValueError(msg)
+
+        if font is None:
+            font = self._getfont(font_size)
 
         if self._multiline_check(text):
             return self.multiline_textbbox(
@@ -561,8 +717,6 @@ class ImageDraw:
                 embedded_color,
             )
 
-        if font is None:
-            font = self.getfont()
         mode = "RGBA" if embedded_color else self.fontmode
         bbox = font.getbbox(
             text, mode, direction, features, language, stroke_width, anchor
@@ -582,23 +736,29 @@ class ImageDraw:
         language=None,
         stroke_width=0,
         embedded_color=False,
+        *,
+        font_size=None,
     ):
         if direction == "ttb":
-            raise ValueError("ttb direction is unsupported for multiline text")
+            msg = "ttb direction is unsupported for multiline text"
+            raise ValueError(msg)
 
         if anchor is None:
             anchor = "la"
         elif len(anchor) != 2:
-            raise ValueError("anchor must be a 2 character string")
+            msg = "anchor must be a 2 character string"
+            raise ValueError(msg)
         elif anchor[1] in "tb":
-            raise ValueError("anchor not supported for multiline text")
+            msg = "anchor not supported for multiline text"
+            raise ValueError(msg)
+
+        if font is None:
+            font = self._getfont(font_size)
 
         widths = []
         max_width = 0
         lines = self._multiline_split(text)
-        line_spacing = (
-            self.textsize("A", font=font, stroke_width=stroke_width)[1] + spacing
-        )
+        line_spacing = self._multiline_spacing(font, spacing, stroke_width)
         for line in lines:
             line_width = self.textlength(
                 line,
@@ -637,7 +797,8 @@ class ImageDraw:
             elif align == "right":
                 left += width_difference
             else:
-                raise ValueError('align must be "left", "center" or "right"')
+                msg = 'align must be "left", "center" or "right"'
+                raise ValueError(msg)
 
             bbox_line = self.textbbox(
                 (left, top),
@@ -749,8 +910,8 @@ def floodfill(image, xy, value, border=None, thresh=0):
     full_edge = set()
     while edge:
         new_edge = set()
-        for (x, y) in edge:  # 4 adjacent method
-            for (s, t) in ((x + 1, y), (x - 1, y), (x, y + 1), (x, y - 1)):
+        for x, y in edge:  # 4 adjacent method
+            for s, t in ((x + 1, y), (x - 1, y), (x, y + 1), (x, y - 1)):
                 # If already processed, or if a coordinate is negative, skip
                 if (s, t) in full_edge or s < 0 or t < 0:
                     continue
@@ -763,7 +924,7 @@ def floodfill(image, xy, value, border=None, thresh=0):
                     if border is None:
                         fill = _color_diff(p, background) <= thresh
                     else:
-                        fill = p != value and p != border
+                        fill = p not in (value, border)
                     if fill:
                         pixel[s, t] = value
                         new_edge.add((s, t))
@@ -813,38 +974,44 @@ def _compute_regular_polygon_vertices(bounding_circle, n_sides, rotation):
     # 1. Error Handling
     # 1.1 Check `n_sides` has an appropriate value
     if not isinstance(n_sides, int):
-        raise TypeError("n_sides should be an int")
+        msg = "n_sides should be an int"
+        raise TypeError(msg)
     if n_sides < 3:
-        raise ValueError("n_sides should be an int > 2")
+        msg = "n_sides should be an int > 2"
+        raise ValueError(msg)
 
     # 1.2 Check `bounding_circle` has an appropriate value
     if not isinstance(bounding_circle, (list, tuple)):
-        raise TypeError("bounding_circle should be a tuple")
+        msg = "bounding_circle should be a tuple"
+        raise TypeError(msg)
 
     if len(bounding_circle) == 3:
         *centroid, polygon_radius = bounding_circle
     elif len(bounding_circle) == 2:
         centroid, polygon_radius = bounding_circle
     else:
-        raise ValueError(
+        msg = (
             "bounding_circle should contain 2D coordinates "
             "and a radius (e.g. (x, y, r) or ((x, y), r) )"
         )
+        raise ValueError(msg)
 
     if not all(isinstance(i, (int, float)) for i in (*centroid, polygon_radius)):
-        raise ValueError("bounding_circle should only contain numeric data")
+        msg = "bounding_circle should only contain numeric data"
+        raise ValueError(msg)
 
     if not len(centroid) == 2:
-        raise ValueError(
-            "bounding_circle centre should contain 2D coordinates (e.g. (x, y))"
-        )
+        msg = "bounding_circle centre should contain 2D coordinates (e.g. (x, y))"
+        raise ValueError(msg)
 
     if polygon_radius <= 0:
-        raise ValueError("bounding_circle radius should be > 0")
+        msg = "bounding_circle radius should be > 0"
+        raise ValueError(msg)
 
     # 1.3 Check `rotation` has an appropriate value
     if not isinstance(rotation, (int, float)):
-        raise ValueError("rotation should be an int or float")
+        msg = "rotation should be an int or float"
+        raise ValueError(msg)
 
     # 2. Define Helper Functions
     def _apply_rotation(point, degrees, centroid):
@@ -893,6 +1060,6 @@ def _color_diff(color1, color2):
     Uses 1-norm distance to calculate difference between two values.
     """
     if isinstance(color2, tuple):
-        return sum([abs(color1[i] - color2[i]) for i in range(0, len(color2))])
+        return sum(abs(color1[i] - color2[i]) for i in range(0, len(color2)))
     else:
         return abs(color1 - color2)
